{% extends 'base.html.j2' %}
{% set page_title = entity.label | localize %}
{% set entity_contexts = entity_contexts(entity) %}
{% do breadcrumbs.append(
    entity_type.entity_type_label_plural() | localize,
    entity_type | entity_type_name | camel_case_to_kebab_case | url(absolute=true)
) %}
{% do breadcrumbs.append(
    page_title,
    page_resource | url(absolute=true)
) %}
{% block page_content %}
{% endblock %}
